
/*  trainEigenFloat.c  - July '95 */
/*  VERSION 1.0:
    Input sequence of float training images,
     compute SVD,
     output singular values and leading principal components.

     Usage:
      trainEigen fileSeq.*.pfm
     Expects spatially register float (pfm) images all of the same size.

     Writes float principal components to pathname_eigNUM.pfm
            byte      ditto             pathname_eigNUM.pgm
     Singular values written in pathname_eig.sv
  
******************/

# include   "macros.h"
# include   "utils.h"
# include   "imageFile-io.h"
# include "svd.h"

unsigned char grey = 127;

int main(int argc, char **argv)
{
  float *work, w[SVD_DIM], rv1[SVD_DIM], *v, *imageFlt; 
  float *A, *meanImage;
  float tmp, mean, var, stdev;
  float scl, meanMean, sqrLengthMean;
  float crop[2], out_of_range[2], quant_par[2];
  float frac, sum;
  unsigned char *image;

  int use_transpose, count, use_mean;
  int svPerm[SVD_DIM];
  int nx0, ny0, nxImage, nyImage, nyWork;
  int nSV, nImage, col;
  int iShift, iOrient;
  int iShiftImage, iShiftWork, c;
  int skip, numImage, pad;
  int ixy, ix, iy, size, i,k, kImage,start;
  int nBasis;
  char fnCurr[MAXLEN], pathName[MAXLEN], rootName[MAXLEN];
  char inName[SVD_DIM][MAXLEN], outName[MAXLEN];
  char *inpath, *outpath, numString[MAXLEN],resp[MAXLEN];
  FILE *fpIn;
  FILE *fpOut;

  nImage = argc-1;
  if (nImage == 0) {
    fprintf(stderr,"usage: trainEigenFloat {imageFile*.pfm}\n");
    exit(0);
  }
  fprintf(stderr," image sequence length = %d\n", nImage);

  if ( nImage > SVD_DIM)
    error(" Sequence length too large, recompile with larger SVD_DIM","");

  fprintf(stderr," enter output file pathname and rootname : ");
  scanf("%s %s",pathName, rootName);

  fprintf(stderr," enter max number of eigen basis functions to output : ");
  scanf("%d", &nBasis);
  nBasis = min(nBasis, nImage);
  fprintf(stderr," compute mean image (and difference from mean): (y/n): ");
  scanf("%s", resp);
  use_mean = TRUE;
  if (resp[0] == 'n' || resp[0] == 'N') {
    use_mean = FALSE;
    fprintf(stderr, " Not using mean\n");
  } else {
    fprintf(stderr, " Using mean\n");
  }

  /* Input sequence, store in work array */
  fpIn = NULL;
  for(kImage=0, col=1; kImage<nImage; kImage++, col++) {
    strcpy(inName[kImage], *++argv);

    /* Open image file */
    if ((fpIn = freopenInputFile(inName[kImage], fpIn)) == NULL) {
      fprintf(stderr, "Can't open image file %s\n",inName[kImage]);
      exit(0);
    }
    read_pfm_image(&imageFlt, fpIn, inName[kImage], &nxImage, &nyImage); 
    if (kImage == 0) {
      nx0 = nxImage;
      ny0 = nyImage;

      size = (nxImage * nyImage +1) * SVD_DIM;
      grabFloatMemory( &work, size, "work");
      grabFloatMemory( &v, size, "v");
    } else {
      if ((nxImage != nx0) || (nyImage !=ny0)) {
        fprintf(stderr, "Error: %s and %s different sizes\n",
		inName[0], inName[kImage]);
	error("Images in input sequence are of different sizes","");
      }
    }

    /* Remove the mean brightness from each training image,
       And scale the pixel variance to be constant (i.e. 1.0 per pixel).  */ 
    var = mean = 0.0;
    for(i=0; i<nxImage*nyImage; i++) {
      mean += imageFlt[i];
      tmp = (imageFlt[i] - grey);
      var += tmp * tmp;
    }
    mean /= nxImage * nyImage;
    var /= nxImage * nyImage;
    var -= (mean - grey) * (mean - grey);
    if (var <= 0) {
      fprintf(stderr,"Warning: Training image variance = %e\n", var);
      stdev = 1.0;
    } else {
      stdev = sqrt(var);
    }
    for(i=0; i<nxImage*nyImage; i++)
      imageFlt[i] = (imageFlt[i] - mean)/stdev;
    

    /* Copy imageFlt to column number col of SVD array work,
       remembering to start at row 1, 1 of the array work */
    for(iy=0; iy<nyImage; iy++) 
      for(ix=0; ix<nxImage; ix++) {
        ixy = index(ix, iy, nxImage);
        work[index(col, ixy+1, SVD_DIM)] = imageFlt[ixy] ;
      }
    free(imageFlt);
  }

  /*  Subtract mean scaled image */
  if (use_mean) {
    grabFloatMemory(&meanImage, nxImage * nyImage, "meanImage");
    for(ixy=0; ixy<nxImage*nyImage; ixy++)
      meanImage[ixy] = 0.0;
    for(ixy=0; ixy<nxImage*nyImage; ixy++) {
      for(col=1; col<=nImage; col++) {
	meanImage[ixy] += work[index(col, ixy+1, SVD_DIM)];
      }
      meanImage[ixy] /= nImage;
    }
    for(ixy=0; ixy<nxImage*nyImage; ixy++) {
      for(col=1; col<=nImage; col++) {
	work[index(col, ixy+1, SVD_DIM)] -= meanImage[ixy];
      }
    }
  }

  if (nxImage*nyImage < nImage) {
    use_transpose = 1;
    size = (nImage + 1) * SVD_DIM;
    grabFloatMemory( &A, size, "A");
    grabFloatMemory( &v, size, "v");
    for(c=1; c <= nImage; c++)
      for(ixy=1; ixy <=nxImage*nyImage; ixy++)
	A[ixy + c * SVD_DIM] = work[ixy*SVD_DIM + c];
    if (svdEispack(A, nImage, nxImage*nyImage, w, v, rv1) < 0)
      error(" SVD computation dies hard.","");
    sortSV(svPerm, w, nxImage*nyImage);
    nSV = nxImage * nyImage;

  } else {
    use_transpose = 0;
    size = (nxImage * nyImage + 1) * SVD_DIM;
    grabFloatMemory( &A, size, "A");
    grabFloatMemory( &v, size, "v");
    for(c=1; c <= nImage; c++)
      for(ixy=1; ixy <=nxImage*nyImage; ixy++)
	A[ixy*SVD_DIM + c] = work[ixy*SVD_DIM + c];
    if (svdEispack(A, nxImage*nyImage, nImage, w, v, rv1) < 0)
      error(" SVD computation dies hard.","");
    sortSV(svPerm, w, nImage);
    nSV = nImage;
  }

  /* Check out the mean and power of mean image */
  meanMean = 0.0;
  sqrLengthMean = 0.0;
  for(ixy=0; ixy<nxImage*nyImage; ixy++) {
      tmp = meanImage[ixy];
      meanMean += tmp;
      sqrLengthMean += tmp * tmp;
  }
  meanMean /= nxImage * nyImage;
  sqrLengthMean = sqrLengthMean - meanMean * meanMean * nxImage * nyImage;
  fprintf(stderr, " meanMean %e sqrLengthMean %e\n", meanMean, sqrLengthMean);
  /* The mean of the mean image should be nearly zero... make it zero,
     and scale it so it has length one. */
  if (sqrLengthMean <= 0.0) {
     sqrLengthMean = 0.0;
     fprintf(stderr,"Warning: DC mean image.\n");
     use_mean = FALSE;
  } else {
    /* Scale mean image to have unit length */
    scl = 1.0/sqrt(sqrLengthMean);
    for(ixy=0; ixy<nxImage*nyImage; ixy++)
      meanImage[ixy] = (meanImage[ixy] - meanMean) * scl;
    /* Scale singular values by the same factor */
    for(k=1; k<=nSV; k++)
      w[k] *= scl;
  }

  /* DEBUG Check the mean image has zero DC brightness, and is
     of unit length. */
  meanMean = 0.0;
  sqrLengthMean = 0.0;
  for(ixy=0; ixy<nxImage*nyImage; ixy++) {
      tmp = meanImage[ixy];
      meanMean += tmp;
      sqrLengthMean += tmp * tmp;
  }
  meanMean /= nxImage * nyImage;
  sqrLengthMean = sqrLengthMean - meanMean * meanMean * nxImage * nyImage;
  fprintf(stderr, "After scaling: meanMean %e sqrLengthMean %e\n", 
	  meanMean, sqrLengthMean);
 

  /* Output ASCII file with singular values */
  fpOut = NULL;
  strcpy(fnCurr,pathName);
  strcat(fnCurr,rootName);
  strcat(fnCurr,"_eig.sv");
  if ((fpOut = fopen(fnCurr, "w")) == NULL) 
    fprintf(stderr, "Cannot open singular value output file %s\n", fnCurr);
  else {
    for(k=1; k<=nSV; k++) {
      fprintf(fpOut, "%e\n", w[svPerm[k]]);
    }
  }

  /* Put mean image into first column of the U/V matrix,
     depending on whether or not we did svd on the transpose of
     the training images or not.  */
  svPerm[0] = 0;
  if (use_mean) {
    if (use_transpose)
      for(ixy=0; ixy<nxImage * nyImage; ixy++)
        v[index(0, ixy+1, SVD_DIM)] = meanImage[ixy];
    else
      for(ixy=0; ixy<nxImage * nyImage; ixy++)
        A[index(0, ixy+1, SVD_DIM)] = meanImage[ixy];
  } else {
    if (use_transpose)
      for(ixy=0; ixy<nxImage * nyImage; ixy++)
        v[index(0, ixy+1, SVD_DIM)] = 0.0;
    else
      for(ixy=0; ixy<nxImage * nyImage; ixy++)
        A[index(0, ixy+1, SVD_DIM)] = 0.0;
  }

  /** Write out principal component images  and mean image (0 if not used) **/
  grabFloatMemory( &imageFlt, nxImage*nyImage, "imageFloat");
  for(kImage=0; kImage<=nBasis; kImage++) {

    k = svPerm[kImage];
    if (use_transpose > 0) {
      for(i=0; i<nxImage*nyImage; i++) {
	ixy = index(k, (i+1), SVD_DIM);  /* column k, row i+1 */
	imageFlt[i] = v[ixy];
      }
    } else {
      for(i=0; i<nxImage*nyImage; i++) {
	ixy = index(k, (i+1), SVD_DIM);  /* column k, row i+1 */
	imageFlt[i] = A[ixy];
      }
    }
    if (!use_mean && kImage == 1) { /* Check sign of first principal vector */
      count = 0; 
      for(c = 0; c<nImage; c++) {
	sum = 0.0;
	for(i=0; i<nxImage*nyImage; i++) 
	  sum += imageFlt[i] * work[index(c+1, i+1, SVD_DIM)];
	if (sum > 0.0)
	  count++;
      }
      frac = count/(float)nImage;
      if (frac < 0.5) {
	for(i=0; i<nxImage*nyImage; i++) 
	  imageFlt[i] = -imageFlt[i];
	frac = 1.0 - frac;
      }
      fprintf(stderr, " Check sign gets %f percent positive\n", 100.0 * frac);
    }

    /* Save float eigen image */
    strcpy(fnCurr,pathName);
    strcat(fnCurr,rootName);
    strcat(fnCurr,"_eig");
    pad_integer(numString, kImage, 3);
    sprintf(outName, "%s%s_eig%s",pathName, rootName, numString);
    save_pfm_image(outName, imageFlt, nxImage, nyImage);

    /* Save pgm form eigen image, scaled symmetrically about zero */
    range_float_image(imageFlt, nxImage, nyImage, crop);
    out_of_range[0] = 0.0; out_of_range[1] = 1.0;
    tmp = max(abs(crop[0]), abs(crop[1]));
    if (tmp <= 0.0 && (kImage>0 || use_mean)) {
      fprintf(stderr, "Warning: zero eigen image.\n");
      tmp = 1.0;
    }
    crop[1] = tmp;
    crop[0] = -tmp;
    quantize_crop_range_float_image(&image, imageFlt, nxImage, nyImage,
				    crop, out_of_range, quant_par);
    save_pgm_image(outName, image, nxImage, nyImage);

    free(image);
  }
  free(imageFlt);
  free(work);
  free(v);
  return(0);
}


